#!/bin/sh -e
# ==============================================================================
# portsreinstall library script
# - Operations for queries to the temporary database -
# Copyright (C) 2013 Mamoru Sakaue, MwGhennndo, All Rights Reserved.
# This software is distributed under the 2-Clause BSD License.
# ==============================================================================

# ============= Show a list of failed ports with their reasons =============
database_query_show_list_failure ()
{
	if [ `cat "${DBDIR}/failed.list" 2> /dev/null | wc -l` -eq 0 ]
	then
		message_echo "INFO: No item is registered in this list."
		return 1
	fi
	grep -v -E -f "${DBDIR}/ports_to_delete.grep_pattern" "${DBDIR}/failed.list" \
		| while read origin
	do
		origin_regexp=`str_escape_regexp "$origin"`
		grep -m 1 -E "^$origin_regexp$" "${DBDIR}/stage.loop_list/ports_to_delete" \
			2> /dev/null && continue
		note=`cat "${DBDIR}/requires/$origin/note_failtre"`
		resolved=no
		grep -m 1 -E "^`str_escape_regexp \"$origin\"`$" \
			"${DBDIR}/manually_done.list" > /dev/null 2>&1 \
			&& resolved=yes
		pkgtag=`cat "${DBDIR}/requires/$origin/pkgtag" 2> /dev/null` || :
		if [ $opt_batch_mode = no ]
		then
			case $resolved in
			no)	resolved=;;
			yes)	resolved=', resolved';;
			esac
			if [ -n "$pkgtag" ]
			then
				echo "$origin ($pkgtag) (error while [$note]$resolved)"
			else
				echo "$origin (error while [$note]$resolved)"
			fi
		else
			printf "%s\t%s\t%s\t%s\n" "$origin" "$pkgtag" "$note" "$resolved"
		fi
	done
	:
}

# ============= Show a list of failed restoration of conflicts =============
database_query_show_list_failed_conflicts_restoration ()
{
	if [ `cat "${DBDIR}/deleted_conflicts" 2> /dev/null | wc -l` -eq 0 ]
	then
		message_echo "INFO: No item is registered in this list."
		return 1
	fi
	grep -v -E -f "${DBDIR}/ports_to_delete.grep_pattern_col1" "${DBDIR}/deleted_conflicts" \
		| if [ $opt_batch_mode = no ]
		then
			while read origin pkg
			do
				if [ -n "$pkg" ]
				then
					echo "$origin ($pkg)"
				else
					echo "$origin"
				fi
			done
		else
			cat
		fi
	:
}

# ============= Insert initial origins to a list of origins =============
database_query_add_initial_origins ()
{
	local origin
	while read origin
	do
		echo "$origin"
		[ -e "${DBDIR}/requires/$origin/initial_orig" ] || continue
		cat "${DBDIR}/requires/$origin/initial_orig"
	done
	:
}

# ============= Get target attributes =============
database_query_get_target_attributes ()
{
	local prefix origin _is_all _is_target _is_requires_requirements _is_initial_requirements _is_requires_dependents _is_initial_dependents _is_requires_requirements_complement _is_relevant infofile tag level
	prefix=$1
	origin=$2
	_is_all=y
	_is_target=
	_is_requires_requirements=
	_is_initial_requirements=
	_is_requires_dependents=
	_is_initial_dependents=
	_is_requires_requirements_complement=
	_is_relevant=y
	if [ ! -e "${DBDIR}/target_all" -a -n "$opt_target_itself$opt_target_dependents$opt_target_requirements" ]
	then
		_is_all=
		tag=`options_get_dependency_type`
		level=full
		[ $opt_only_target_scope = yes ] && level=direct
		infofile=${DBDIR}/targets/$origin/attrs.${tag}.${level}
		if [ -e "$infofile" ]
		then
			. "$infofile"
		else
			_is_relevant=
		fi
	fi
	eval ${prefix}_is_all=\$\{_is_all\}
	eval ${prefix}_is_target=\$\{_is_target\}
	eval ${prefix}_is_requires_requirements=\$\{_is_requires_requirements\}
	eval ${prefix}_is_initial_requirements=\$\{_is_initial_requirements\}
	eval ${prefix}_is_requires_dependents=\$\{_is_requires_dependents\}
	eval ${prefix}_is_initial_dependents=\$\{_is_initial_dependents\}
	eval ${prefix}_is_requires_requirements_complement=\$\{_is_requires_requirements_complement\}
	eval ${prefix}_is_relevant=\$\{_is_relevant\}
}

# ============= Check whether (re/de)installation of a port is suppressed =============
database_query_is_a_port_suppressed ()
{
	local origin flag
	origin=$1
	if [ $opt_suppress_self_upadte = yes ]
	then
		flag=SUPPRESSED_SELF
	elif [ $opt_suppress_pkgtools_upadte = yes ]
	then
		flag=SUPPRESSED_PKGNG
	else
		return 1
	fi
	for db in initial requires
	do
		[ -e "${DBDIR}/$db/$origin/$flag" ] && return
	done
	return 1
}

# ============= Check whether a port needs to be updated or upgraded =============
database_query_does_a_port_need_update ()
{
	local origin dbpath
	origin=$1
	dbpath=${DBDIR}/requires/$origin
	[ -e "$dbpath/conf_updated" ] && return
	[ -e "$dbpath/new_version" ] || return
	! diff "$dbpath/new_version" "$dbpath/current_version" > /dev/null 2>&1
}

# ============= Check before operations of a command which need the temporary database completely prepared =============
database_query_chk_preparation_completion ()
{
	program_chk_stage_complete PREPARATION && return
	message_echo "ERROR: Database has to be built completely before executing this command." >&2
	exit 1
}

# ============= Get a make variable value of a port =============
database_query_get_makevar_val ()
{
	local origin variable dbdir cache value
	origin=$1
	variable=$2
	dbdir=${DBDIR}/requires/$origin
	cache=$dbdir/makevar/$variable
	if [ -e "$cache" ]
	then
		cat "$cache"
	else
		value=`make -C "${PORTSDIR}/$origin" -V "$variable"`
		if [ -d "$dbdir" ] && misc_is_superuser_privilege
		then
			[ -d "$dbdir/makevar" ] || mkdir "$dbdir/makevar"
			echo "$value" > $cache.tmp
			mv "$cache.tmp" "$cache"
		fi
		echo "$value"
	fi
}

# ============= Check whether configurations for a port is default =============
database_query_is_default_conf ()
{
	local origin mode dbpath tmp_msg is_customized is_requiremnt_replaced files origin_regexp tmp_old tmp_new origin_requirement
	origin=$1
	mode=$2
	dbpath=${DBDIR}/requires/$origin
	if [ ! -e "$dbpath/is_customized" ]
	then
		tmp_msg=${TMPDIR}/database_query_is_default_conf:msg
		cp /dev/null "$tmp_msg"
		is_customized=no
		if [ `ls "${DBDIR}/conf/each_port/$origin" 2> /dev/null | wc -l` -gt 0 ]
		then
			files=`ls "${DBDIR}/conf/each_port/$origin" | str_cancat_items_for_sentence`
			echo "Knobs and miscellaneous customization by $files," >> $tmp_msg
			is_customized=yes
		fi
		if ! diff "$dbpath/ports_options.default" "$dbpath/ports_options.current" > /dev/null 2>&1
		then
			echo "Non-default port options," >> $tmp_msg
			is_customized=yes
		fi
		origin_regexp=`str_escape_regexp "$origin"`
		if grep -m 1 -E "^$origin_regexp$" "${DBDIR}/conf/NOPKG:PORTS.parsed" > /dev/null 2>&1
		then
			echo "Explicit specification as non-default in ${APPNAME}.conf," >> $tmp_msg
			is_customized=yes
		fi
		tmp_old=${TMPDIR}/database_query_is_default_conf:old
		tmp_new=${TMPDIR}/database_query_is_default_conf:new
		is_requiremnt_replaced=no
		if fileedit_manipulate_old_new_lines "$dbpath/requirements.all.direct.orig" "$dbpath/requirements.all.direct" \
			"$tmp_old" "$tmp_new"
		then
			echo "Replacement in requirements:" >> $tmp_msg
			echo "-------- FROM --------" >> $tmp_msg
			sed 's/^/ /' "$tmp_old" >> $tmp_msg
			echo "--------  TO  --------" >> $tmp_msg
			sed 's/^/ /' "$tmp_new" >> $tmp_msg
			echo "----------------------" >> $tmp_msg
			is_customized=yes
		fi
		for origin_requirement in `cat "$dbpath/requirements.all.direct" 2> /dev/null`
		do
			database_query_is_default_conf "$origin_requirement" quiet && continue
			echo "Non-default requirement $origin_requirement," >> $tmp_msg
			is_customized=yes
		done
		{ [ $is_customized = yes ] && cat "$tmp_msg"; } > $dbpath/is_customized.tmp
		mv "$dbpath/is_customized.tmp" "$dbpath/is_customized"
	fi
	[ `wc -c < $dbpath/is_customized` -eq 0 ] && return
	if [ "x$mode" != xquiet ]
	then
		message_echo "INFO: This port is configured to be non-default because of"
		message_cat 3<< eof
`sed 's/^/         /' "$dbpath/is_customized"`
eof
		message_echo "      so the prebuilt package is not used."
	fi
	return 1
}

# ============= Output of "show" command for each matching port =============
database_query_for_each_matching_port ()
{
	local grandtitle title list isfirst origin_target pkg_target table_target
	grandtitle=$1
	title=$2
	list=$3
	pkgnamedb=$4
	deptag=$5
	level=$6
	shift 6
	message_echo "[$grandtitle]"
	message_dependency_scope
	message_echo
	isfirst=y
	for origin_target in `pkgsys_eval_ports_glob "$@"`
	do
		pkg_target=
		for table_target in $pkgnamedb
		do
			pkg_target=`cat "${DBDIR}/$table_target/$origin_target/pkgtag" 2> /dev/null` || :
			[ -n "$pkg_target" ] && break
		done
		[ -n "$pkg_target" ] || continue
		[ "$isfirst" = y ] || message_echo
		isfirst=n
		[ $opt_batch_mode = no ] && printf "$title\n" "$origin_target ($pkg_target)"
		list_target=
		for table_target in $pkgnamedb
		do
			list_target=${DBDIR}/$table_target/$origin_target/$list
			[ -e "$list_target" ] && break
		done
		[ -e "$list_target" ] || continue
		if [ $opt_batch_mode = no ]
		then
			while read origin
			do
				pkg=
				for table in $pkgnamedb
				do
					pkg=`cat "${DBDIR}/$table/$origin/pkgtag" 2> /dev/null` || :
					[ -n "$pkg" ] && break
				done
				[ -n "$pkg" ] || continue
				echo "$origin ($pkg)"
			done < $list_target
		else
			while read origin
			do
				pkg=
				for table in $pkgnamedb
				do
					pkg=`cat "${DBDIR}/$table/$origin/pkgtag" 2> /dev/null` || :
					[ -n "$pkg" ] && break
				done
				[ -n "$pkg" ] || continue
				printf '%s\t%s\t%s\t%s\n' "$origin_target" "$pkg_target" "$origin" "$pkg"
			done < $list_target
		fi
	done
	if [ "$isfirst" = y ]
	then
		message_echo "ERROR: No matching port for the glob(s) is in the temporary database." >&2
		exit 1
	fi
	:
}

# ============= Output of "show" command for a single list =============
database_query_show_single_list ()
{
	local list pkgnamedb flag_filter_skip_unchanged flag_filter_only_target tmpflag_exists put_blankline
	list=$1
	pkgnamedb=$2
	flag_filter_skip_unchanged=$3
	flag_filter_only_target=$4
	tmpflag_exists=${TMPDIR}/database_query_show_single_list::exists_item
	if [ `cat "${DBDIR}/$list" 2> /dev/null | wc -l` -eq 0 ]
	then
		message_echo "INFO: No item is registered in this list."
		return 1
	fi
	rm -f "$tmpflag_exists"
	put_blankline=
	if [ -n "$flag_filter_only_target" \
		-a -n "$opt_target_itself$opt_target_dependents$opt_target_requirements" ]
	then
		message_echo "INFO: Ports outside of the target scope are excluded."
		put_blankline=y
	fi
	if [ -n "$flag_filter_skip_unchanged" -a $opt_skip_unchanged = yes ]
	then
		message_echo "INFO: Ports which have been the newest with their all requirements from the first are excluded."
		put_blankline=y
	fi
	[ -n "$put_blankline" ] && message_echo
	while read origin
	do
		[ -n "$flag_filter_skip_unchanged" -a $opt_skip_unchanged = yes \
			-a ! -e "${DBDIR}/requires/$origin/$flag_filter_skip_unchanged" ] \
			&& continue
		if [ -n "$flag_filter_only_target" ]
		then
			database_query_get_target_attributes currentorigin "$origin"
			[ -n "${currentorigin_is_relevant}" ] || continue
			database_query_is_a_port_suppressed "$origin" && continue
		fi
		pkg=
		for table in $pkgnamedb
		do
			pkg=`cat "${DBDIR}/$table/$origin/pkgtag" 2> /dev/null` || :
			[ -n "$pkg" ] && break
		done
		[ -n "$pkg" ] || continue
		touch "$tmpflag_exists"
		if [ $opt_batch_mode = no ]
		then
			echo "$origin ($pkg)"
		else
			printf '%s\t%s\n' "$origin" "$pkg"
		fi
	done < ${DBDIR}/$list
	[ -e "$tmpflag_exists" ] && return
	message_echo "INFO: No item is registered in this list."
	return 1
}

# ============= Check whether the upgrade is necessary for a port =============
database_query_is_necessary_upgrade ()
{
	local origin nodedir dbsuffix tmpfile_new tmpfile_old tmpfile_diff
	origin=$1
	nodedir=${DBDIR}/requires/$origin
	[ -e "$nodedir/installed_by_pkg" ] && return 1
	dbsuffix=`options_get_dependency_type`.`options_get_dependency_level`
	tmpfile_new=${TMPDIR}/database_query_is_necessary_upgrade:failed_requirements.new
	tmpfile_old=${TMPDIR}/database_query_is_necessary_upgrade:failed_requirements.old
	tmpfile_diff=${TMPDIR}/database_query_is_necessary_upgrade:failed_requirements.diff
	[ -e "$nodedir/failed_requirements.${dbsuffix}.previous" ] || return
	sort -u "$nodedir/failed_requirements.${dbsuffix}" > $tmpfile_new 2> /dev/null || :
	sort -u "$nodedir/failed_requirements.${dbsuffix}.previous" > $tmpfile_old 2> /dev/null || :
	fileedit_exists_old_lines "$tmpfile_old" "$tmpfile_new"
}

# ============= Actual operations of "show" command for a single list =============
database_query_show_single_list_exec ()
{
	local subject deptag level dbsuffix flag_filter_skip_unchanged flag_filter_only_target pkgnamedb
	subject=$1
	deptag=$2
	level=$3
	dbsuffix=$deptag.$level
	flag_filter_skip_unchanged=
	flag_filter_only_target=
	pkgnamedb='requires obsolete initial'
	case $subject in
	todo)
		message_echo "The following ports remain in the (re)installation queue for the current do/redo process:"
		message_echo "It is noted that ports to be skipped can be included here."
		message_dependency_scope
		message_echo
		list=stage.loop_list/reinst_todo.remain
		[ ${DBDIR}/reinst_order.list -nt ${DBDIR}/$list ] && list=reinst_order.list
		flag_filter_skip_unchanged=necessary_upgrade.$dbsuffix
		flag_filter_only_target=y
		;;
	done)
		message_echo "The following ports have been successfully (re)installed or newly installed:"
		message_dependency_scope
		message_echo
		list=success.$dbsuffix.list
		flag_filter_skip_unchanged=necessary_upgrade_completed.$dbsuffix
		flag_filter_only_target=y
		;;
	redo)
		message_echo "The following ports need (re)installation but are to be skipped until any of their failed requirements succeeds:"
		message_dependency_scope
		message_echo
		list=todo_after_requirements_succeed.$dbsuffix.list
		flag_filter_skip_unchanged=necessary_upgrade.$dbsuffix
		flag_filter_only_target=y
		;;
	resolved)
		message_echo "The following ports had problems which have been manually resolved:"
		message_echo
		list=manually_done.list
		;;
	failure)
		message_echo "The following ports experienced failures and kept to be old or uninstalled:"
		message_echo
		database_query_show_list_failure
		return
		;;
	conflict)
		message_echo "The following ports are temporarily deleted due to conflicts:"
		message_echo
		database_query_show_list_failed_conflicts_restoration
		return
		;;
	taboo)
		message_echo "The following ports are registered as taboo:"
		message_echo
		list=taboo.all.list
		;;
	need)
		message_echo "The following ports are registered as necessary:"
		message_echo
		list=need.list
		;;
	noneed)
		message_echo "The following ports are registered as unnecessary:"
		message_echo
		list=noneed.list
		;;
	restored)
		message_echo "The following leaf, obsolete or unneeded ports had been once deleted but are to be or have been restored:"
		message_echo
		list=stage.loop_list/ports_to_restore
		pkgnamedb='obsolete initial'
		flag_filter_only_target=y
		;;
	deleted)
		message_echo "The following leaf, obsolete or unneeded ports are to be or have been deleted:"
		message_echo
		list=stage.loop_list/ports_to_delete
		pkgnamedb='obsolete initial'
		flag_filter_only_target=y
		;;
	esac
	database_query_show_single_list "$list" "$pkgnamedb" \
		"$flag_filter_skip_unchanged" "$flag_filter_only_target"
}
